/*	AA script rendering an SVG file, with clickable areas corresponding to a Blocks property.

 	Copyright (c) 2020 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
 */


(function () {	// Wrap the whole shebang in an IIFE to not pollute the global namespace

	/**
	 * Instantiate a PubSubPeer that auto-connects to the blocks server
	 * using a websocket.
	 *
	 * The PIXILAB Websocket API is defined in the pub-sub-peer.js script, which
	 * attaches its root object as a global (window) variable named PIXILAB_BLOCKS.
	 */
	const pubSubPeer:any = new PIXILAB_BLOCKS.PubSubPeer(
		onServerConnectionChange
	);

	const blocksVarName = getQueryParams()['blocksVar'];
	const svgFileName = getQueryParams()['svgFileName'];

	/**
	 * The server connection state has changed. Here we just log that event, but you
	 * may want to use it to indicate the server is temporarily offline, or similar.
	 */
	function onServerConnectionChange(connected: boolean) {
		console.log(connected ? 'Connected to the Blocks server' : 'Disconnected from the Blocks server');
	}

	/**
	 * Get the set of query params specified in URL, building it if not yet done.
	 */
	function getQueryParams(): Dictionary<string> {
		if (!queryParams) {
			const query = window.location.search.substring(1);
			const vars = query.split("&");
			const qParams:  { [param: string]: string; } = {};
			if (vars.length > 0 && vars[0].length > 0) {	// Got some fish
				for (var i = 0; i < vars.length; i++) {
					const pair = vars[i].split('=');
					var value = pair[1];
					value = value.split('?')[0];	// Discard any garbage added by SSSP v2
					qParams[pair[0]] = value;
				}
			}
			queryParams = qParams;
		}
		return queryParams;
	}
	var queryParams: Dictionary<string>;	// Use ONLY through getQueryParams

	// A simple typed dictionary type, always using string as key export
	interface Dictionary<TElem> {
		[id: string]: TElem;
	}

	/**
	 *	Load an external SVG file into the DOM as a SVG element. Listen for
	 *	clicks on children of #TheMap inside the SVG to select clicked
	 *	child, indicated by applying the class 'svg-selected' to the selected child.
	 */
	class SvgMap {
		private svgWrap: Element;	// The wrapping group inside he SVG, once loaded

		/*	Create me by loading SVG from external file.
		 */
		constructor(private addToEl: HTMLElement) {
			fetch(svgFileName).	// Selected in the query parameter svgFileName.
			then(response => {
				if (response.ok)
					return response.text();
				else
					throw response.statusText + ': ' + response.url;
			}).
			then(svgDoc => this.addSvgDoc(svgDoc)).
			catch(error => addToEl.innerText = error); // Provides some error ind

			// const dataHandler = {dataReceived: this.onVariableValueChange};
			// this.onVariableValueChange(pubSubPeer.subscribe(blocksVarName,dataHandler ));
		}

		/**	Turn addTo XML data into an SVG node and add inside el.
		 */
		addSvgDoc(svgDoc: string) {
			// Turn string into elements we can work with
			const elems = SvgMap.parseXml(svgDoc);
			// Look in that collection for the svg element in there
			for (var eix = 0; eix < elems.length; ++eix) {
				var svg:Element = elems[eix];
				if (svg.tagName === 'svg') {
					// Discard attribute not kosher for inline svg
					svg.removeAttribute('xmlns:a');
					this.addSvgEl(<SVGElement>svg);
					return;
				}
			}
		}

		/**
		 * Attempt to intepret string into something DOM-like, returning an HTMLCollection
		 */
		static parseXml(str: string): HTMLCollection {
			var tmp = document.implementation.createHTMLDocument();
			tmp.body.innerHTML = str;
			return tmp.body.children;
		}

		/**
		 * Add the SVG element to the DOM. Look for our "interesting" element group,  attach
		 * a click handler there and start to subscribe to the Blocks variable.
		 */
		addSvgEl(svg: SVGElement) {
			this.addToEl.appendChild(svg);
			const theMap = this.svgWrap = document.getElementById('TheMap');
			if (theMap) {
				theMap.addEventListener('mousedown', evt => {
					this.clickedEl(this.findIdentifiedSvgEl(<Element>evt.target));
				});
				const dataHandler = {dataReceived: blocksData =>
					this.onVariableValueChange(blocksData)
				};
				this.onVariableValueChange(pubSubPeer.subscribe(blocksVarName,dataHandler ));

			}
		}

		/**
		 * Starting from el, find the closest element upwards in the hierarchy
		 * that has an ID, and return it. Return undefined if none found prior
		 * to reaching our wrapper.
		 */
		findIdentifiedSvgEl(el: Element): Element | undefined {
			do {
				if (el.id)	// Found one with ID
					return el;
				el = el.parentElement;	// Continue up the hierarchy
			} while (el !== this.svgWrap);	// Stop once we reach wrapping el
		}

		/**
		 * Things to do if an element has been clicked.
		 */
		clickedEl(el?: Element) {
			pubSubPeer.set(blocksVarName, el ? el.id : ""); //update Blocks variable on new selections

		}

		/**
		 * Things to do when variable is updated in Blocks
		 */
		onVariableValueChange(blocksData :string) {

			if (blocksData !== undefined) {
				let el = document.getElementById(blocksData);
				this.selectEl(el);
			}
		}





		/**
		 * Select el by applying svg-selected class to it. Remove from any
		 * previously selected. If el is undefined, then only remove from
		 * any previously selected.
		 */
		selectEl(el?: Element) {
			if (this.lastSelected && this.lastSelected !== el) {
				this.lastSelected.classList.remove('svg-selected');
				this.lastSelected = undefined;
			}
			if (el) {
				el.classList.add('svg-selected');
				this.lastSelected = el;
			}
			return
		}
		private lastSelected :Element;

	}


	// Get us going
	new SvgMap(document.getElementById('svg'));
})();
